import java.util.*;
import java.lang.*;
import java.io.*;

/**
 * <p>Implementation of the ProofObligationASTExtractor.</p>
 * <p>Read input Oml AST file generated by the Overture Parser,
 * and generate a VDM set with each OmlValueShape node, found in the AST.</p>
 *
 * @author Miguel Ferreira <b>miguel@di.uminho.pt</b>
 * @author Samuel Silva <b>silva.samuel@gmail.com</b>
 * @version 1.0
 */
public class ProofObligationExtractorLogic extends Object
{
	// list to store the diferent Proof Obligations
	private LinkedList<ExtractedProofObligation> po_list = 
		new LinkedList<ExtractedProofObligation>();

	// store the lines of each Proof Obligation
	private StringBuffer po = new StringBuffer();
	// store each Proof Obligation's type
	private String po_type = null;

	// count Proof Obligations
	private int contador = 0;

	// separator prefix
	private String separator = "----------";

	// Proof Obligation type prefixes
	private String map_application = " map application";
	private String fun_application = " function application";
	private String non_empty_sec   = " non-empty sequence";
	private String subtype         = " subtype";
	private String invariant       = " invariants";

	// start of the one line that doesn't matter at all
	// it is stored here because the type definition line is harder to find
	private String nop_line = "Integrity property ";

	// input file name
	private String fileName = null;

	/**
	 * Constructor.
	 */
	ProofObligationExtractorLogic () { }
	
	/**
	 * Process a line obtained in the YYINITIAL state.
	 *
	 * @param line the line to process.
	 * @return true if the processed line is the separator line.
	 */
	public boolean yyinitialLINE(String line) {
		boolean result = false;
		
		// check if line is the seperator
		if(isSeparatorLine(line)) {
			// enter Proof Obligation state because
			// next lines, until the next empty line,
			// are a Proof Obligation
			result = true;
		}
		// check if line is where Proof Obligation's type is specified
		else if(isTypeDefLine(line)) {
			// split the line by ':'
			String[] parts = line.split(":");

			// last part is the one that defines the
			// next Proof Obligation's type
			po_type = parts[parts.length - 1];	// expect that at least line.split(x) = line
		}
		
		return result;
	}
	
	/**
	 * Process a new line obtained in the YYINITAL state.
	 */
	public void inpoNEWLINE() throws java.lang.Exception {
		ProofObligationType type;

		// get it's type
		type = getProofObligationType(po_type);
		
		// store the collected lines in the list
		po_list.addLast(new ExtractedProofObligation(type, po.toString()));

		// clear the po StringBuffer
		po = new StringBuffer();
		// clear type buffer
		po_type = null;
	}
	
	/**
	 * Process a line obtained int he INPO state.
	 *
	 * @param line the line to be processed.
	 */
	public void inpoLINE(String line) {
		// collect the line
		po.append(line);
	}
	
	/**
	 * Process colected data in order to generate a VDM++ class with the 
	 * Proof Obligations as bool values.
	 */
	public void eof() {
		// post processing code

		//System.out.println("Got " + po_list.size() + " Proof Obligations.\n");

		//for(String proofOblg : po_list) {
		//	System.out.println(proofOblg);
		//}

		try {
			// generate VDM++ class
			String classTxt = VppCodeGenerator.printPoValuesClass("PoValues", "SuperClass", po_list, true);

			// print it
			System.out.println(classTxt);
		} catch(java.lang.Exception e) {
			e.printStackTrace();
		}

		if(fileName != null) {
			try {
				//VppCodeGenerator.writeTypesMap(fileName + ".tmap");
				VppCodeGenerator.writeTypesMapToVdmMap(fileName + ".map");
			} catch(java.lang.Exception e) {
				e.printStackTrace();
			}
		}

		return;
	}
	
	/**
	 * Set the input file name.
	 *
	 * @param fileName the input file name.
	 */
	public void setFileName(String fileName) {
		this.fileName = fileName;
	}
	
	/**
	 * Get the input file name.
	 *
	 * @return the input file name.
	 */
	public String getFileName() {
		return fileName;
	}
	
	/**
	 * Check if a line is the separator line, which contains several '-'.
	 *
	 * @param line the line to be checked.
	 * @return true if line is the separator line.
	 */
	private boolean isSeparatorLine(String line) {
		return line.startsWith(separator);
	}

	/**
	 * Check if a line is a Proof Obligation type definition.
	 *
	 * @param line the line to be checked.
	 * @return true if line is a type definition line
	 */
	private boolean isTypeDefLine(String line) {
		return !line.startsWith(nop_line);
	}

	/**
	 * Get a Proof Obligation type.
	 * 
	 * @param str string extracted from the pog file stating the Proof Obligation's type
	 * @return the type of the Proof Obligation
	 */
	private ProofObligationType getProofObligationType(String str) 
		throws java.lang.Exception 
	{
		ProofObligationType type;

		if(str.startsWith(map_application)) {
			type = ProofObligationType.MAP_APPLICATION;
		}
		else if(str.startsWith(fun_application)) {
			type = ProofObligationType.FUN_APPLICATION;
		}
		else if(str.startsWith(non_empty_sec)) {
			type = ProofObligationType.NON_EMPTY_SEC;
		}
		else if(str.startsWith(subtype)) {
			type = ProofObligationType.SUBTYPE;
		}
		else if(str.startsWith(invariant)) {
			type = ProofObligationType.INVARIANT;
		}
		else {
			throw new java.lang.Exception("Unsupported Proof Obligation Type: " + str);
		}

		return type;
	}
}
