import java.lang.String;
import java.lang.StringBuffer;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.Set;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 * Library to support on the VDM++ code generation.
 *
 * @author Miguel Ferreira <b>miguel@di.uminho.pt</b>
 * @author Samuel Silva <b>silva.samuel@gmail.com</b>
 * @version 1.0
 */
public class VppCodeGenerator {

    public static HashMap<String, ProofObligationType> typesMap = null;

	public static StringBuffer generateSet(LinkedList<StringBuffer> elements)
	{
		StringBuffer result = new StringBuffer();
		
		// is true when processing the first element
		boolean first = true;
		
		// VDM set open '{'
		result.append("{ ");
		
		for(StringBuffer sb : elements) {
			if(!first) {
				// write element separator ','
				result.append(", ");
			}
			
			result.append(sb.toString());
			
			first = false;
		}
		
		// VDM set close '}'
		result.append(" }");
		
		return result;
	}
	
    public static StringBuffer generateValue(String access, String type, 
		String name, ExtractedProofObligation extpo) 
	{
    
        StringBuffer sb = new StringBuffer();

    	// add comment with Proof Obligation Type
    	sb.append("// ").append(extpo.getType().toString()).append('\n');

        // access modifier
        sb.append(access).append('\n');

        // name and type
        sb.append(name).append(" : ").append(type).append(" =\n");
        
        // value expression
        sb.append(extpo.getExpression());
        
        // statement terminator
        sb.append(";\n");


        return sb;
    }

    public static StringBuffer generateValuesBlock(StringBuffer[] values) 
	{
    
        StringBuffer sb = new StringBuffer();

        // block start statement
        sb.append("values\n\n");

        // values
        for(StringBuffer value : values) {
        
            sb.append(value).append('\n');
        }

        return sb;
    }

    public static StringBuffer generateClass(String name, String superClass, 
		StringBuffer[] blocks) 
	{
        
      StringBuffer sb = new StringBuffer();

      // start statement
      sb.append("class ").append(name);
        

        // subclass statement
        if(superClass != null) {
            sb.append(" is subclass of ").append(superClass);
        }
        sb.append("\n\n");

      // blocks
      for(StringBuffer block : blocks) {
      
          sb.append(block).append('\n');
      }

      // end statement
      sb.append("\nend ").append(name);

      return sb;
    }

	public static String printPoValuesClass(String className, String superClass,
		LinkedList<ExtractedProofObligation> pos, boolean storeTypes) 
			throws java.lang.Exception 
  	{
  
    if(storeTypes) {
        // initialize typesMap
        typesMap = new HashMap<String, ProofObligationType>(pos.size());
    }

      String access = "public";
      String namePrefix = "po";
      String type = "bool";

      int counter = 1;
      int i = 0;
      int size = pos.size();

      StringBuffer[] values = new StringBuffer[size];

      // generate values
      for(ExtractedProofObligation po : pos) {
      
        // generate value
        String po_name = namePrefix + counter;
        values[i] = generateValue(access, type, po_name, po);

        if(storeTypes) {
          if(!typesMap.containsKey(po_name)) {
            // add type to map
            typesMap.put(po_name, po.getType());
          }
          else {
            throw new java.lang.Exception("Duplicate name in values definition: '" + po_name + "'");
          }
        }

        // increment values
        counter++;
        i++;
      }

      // generate values block
      StringBuffer block = generateValuesBlock(values);

      // generate class
      StringBuffer classCode = generateClass(className, superClass, 
		new StringBuffer[] { block });

      return classCode.toString();
  }

  	public static void writeTypesMap(String fileName) 
		throws java.io.FileNotFoundException, java.io.IOException 
	{
	FileOutputStream fos = new FileOutputStream(fileName);
	ObjectOutputStream oos = new ObjectOutputStream(fos);
	
	oos.writeObject(typesMap);
	oos.close();
	}

  	public static void readTypesMap(String fileName) 
		throws java.io.FileNotFoundException, java.io.IOException, 
			   java.lang.ClassNotFoundException
	{
		FileInputStream fis = new FileInputStream(fileName);
		ObjectInputStream ois = new ObjectInputStream(fis);
	
		typesMap = (HashMap<String, ProofObligationType>) ois.readObject();
    }

	public static void writeTypesMapToVdmMap(String fileName) 
		throws java.io.FileNotFoundException, java.io.IOException 
	{
		if(typesMap != null) {
			StringBuffer sb = new StringBuffer();
			
			// true while processing the first pair (key,vallue)
			boolean first = true;
			
			// open map
			sb.append("{ ");
			
			// get keys
			Set<String> keys = typesMap.keySet();
			
			// write pair (key,value) as VDM mapping elements
			for(String key : keys) {
				
				// write element separator ','
				if(!first) {
					sb.append(", ");
				}
				
				// write mapping domain element
				sb.append('"').append(key).append('"');
				
				// write mapping domain and range separator
				sb.append(" |-> ");
				
				// get element mapped by key
				ProofObligationType poType = 
					(ProofObligationType) typesMap.get(key);
				
				// write mapping range element
				sb.append(printType(poType));
				
				first = false;
			}
			
			// close map
			sb.append(" }");
			
			// write to file
			FileOutputStream fos = new FileOutputStream(fileName);
			fos.write(sb.toString().getBytes());
			fos.close();
		}
	}

	
	private static String printType(ProofObligationType type) {
		String result = null;
		
		switch(type) {
			case MAP_APPLICATION:
				result = "<MapApplication>";
				break;
			case FUN_APPLICATION:
				result = "<FunctionApplication>";
				break;
			case NON_EMPTY_SEC:
				result = "<NonEmptySeq>";
				break;
			case SUBTYPE:
				result = "<Subtype>";
				break;
			case INVARIANT:
				result = "<Invariants>";
				break;
			default:
				break;
		}
		
		return result;
	}


}